# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

"""Generic deep update function for nested dictionaries.

Seems to be solved already in various ways (do we like an extra dependency for pydantic.deep_update?)
https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
But for snippets, license restrictions exist:
https://www.ictrecht.nl/en/blog/what-is-the-license-status-of-stackoverflow-code-snippets
"""  # noqa: E501

from collections.abc import Iterable, Mapping

from box import Box


def deep_merge(x, y, check_enum=False):
    if type(x) != type(y) and not (isinstance(x, Mapping) and isinstance(y, Mapping)):  # noqa: E721
        raise TypeError(f"deep_merge failed due to type mismatch '{x}' (type: {type(x)}) vs. '{y}' (type: {type(y)})")

    if isinstance(x, Box):
        z = Box(_deep_merge_dict(x, y), frozen_box=True)
    elif isinstance(x, Mapping):
        z = _deep_merge_dict(x, y)
    elif isinstance(x, Iterable) and not isinstance(x, str):
        z = _deep_merge_iterable(x, y)
    else:
        z = y  # update

    return z


def _deep_merge_iterable(x: Iterable, y: Iterable):
    # Merging lists is not trivial.
    # There are a few different strategies: replace, unique, append, intersection, ...
    return y
    # return [*x, *y]
    # return [deep_merge(a, b) for a, b in zip(x, y)]


def _deep_merge_dict(x: Mapping, y: Mapping):
    z = dict(x)
    for key in x:
        if key in y:
            if y[key] is None:
                # this is a special case but we have no other way in yaml to express
                # the deletion of fields from a dictionary in an override config
                del z[key]
            else:
                z[key] = deep_merge(x[key], y[key])
    z.update({key: value for key, value in y.items() if key not in z})
    return z



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import enum


class PandasBackend(str, enum.Enum):
    NUMPY = "numpy"
    ARROW = "arrow"


class Dtype:
    """Base class for all data types."""

    def __eq__(self, rhs):
        """Return ``True`` if this dtype is equal to `rhs`."""
        return isinstance(rhs, Dtype) and type(self) is type(rhs)

    def __hash__(self):
        """Return a hash for this dtype."""
        return hash(type(self))

    def __repr__(self):
        """Return a string representation of this dtype."""
        return self.__class__.__name__

    def __str__(self):
        """Return a string representation of this dtype."""
        return self.__repr__()

    @classmethod
    def is_int(cls):
        """Return ``True`` if this dtype is an integer type."""
        return False

    @classmethod
    def is_float(cls):
        """Return ``True`` if this dtype is a float type."""
        return False

    @classmethod
    def is_subtype(cls, rhs):
        """Return ``True`` if this dtype is a subtype of `rhs`.

        For example, ``Int8.is_subtype(Int())`` is ``True``.
        """
        rhs_cls = type(rhs)
        return (cls is rhs_cls) or (rhs_cls is Int and cls.is_int()) or (rhs_cls is Float and cls.is_float())

    @staticmethod
    def from_sql(sql_type) -> "Dtype":
        """Convert a SQL type to a Dtype."""
        import sqlalchemy as sa

        if isinstance(sql_type, sa.SmallInteger):
            return Int16()
        if isinstance(sql_type, sa.BigInteger):
            return Int64()
        if isinstance(sql_type, sa.Integer):
            return Int32()
        if isinstance(sql_type, sa.Float):
            precision = sql_type.precision or 53
            if precision <= 24:
                return Float32()
            return Float64()
        if isinstance(sql_type, sa.Numeric | sa.DECIMAL):
            # Just to be safe, we always use FLOAT64 for fixpoint numbers.
            # Databases are obsessed about fixpoint. However, in dataframes, it
            # is more common to just work with double precision floating point.
            # We see Decimal as subtype of Float. Pydiverse.transform will convert
            # Decimal to Float64 whenever it cannot guarantee semantic correctness
            # otherwise.
            return Float64()
        if isinstance(sql_type, sa.String):
            return String(sql_type.length)
        if isinstance(sql_type, sa.Boolean):
            return Bool()
        if isinstance(sql_type, sa.Date):
            return Date()
        if isinstance(sql_type, sa.Time):
            return Time()
        if isinstance(sql_type, sa.DateTime):
            return Datetime()
        if isinstance(sql_type, sa.Interval):
            return Duration()
        if isinstance(sql_type, sa.ARRAY):
            return List(Dtype.from_sql(sql_type.item_type))
        if isinstance(sql_type, sa.types.NullType):
            return NullType()

        raise TypeError

    @staticmethod
    def from_pandas(pandas_type) -> "Dtype":
        """Convert a pandas type to a Dtype."""
        import numpy as np
        import pandas as pd

        if isinstance(pandas_type, pd.ArrowDtype):
            return Dtype.from_arrow(pandas_type.pyarrow_dtype)

        def is_np_dtype(type_, np_dtype):
            return pd.core.dtypes.common._is_dtype_type(type_, pd.core.dtypes.common.classes(np_dtype))

        workaround = pandas_type is not np.floating  # see https://github.com/pandas-dev/pandas/issues/62018
        if workaround and pd.api.types.is_signed_integer_dtype(pandas_type):
            if is_np_dtype(pandas_type, np.int64):
                return Int64()
            elif is_np_dtype(pandas_type, np.int32):
                return Int32()
            elif is_np_dtype(pandas_type, np.int16):
                return Int16()
            elif is_np_dtype(pandas_type, np.int8):
                return Int8()
            raise TypeError
        if workaround and pd.api.types.is_unsigned_integer_dtype(pandas_type):
            if is_np_dtype(pandas_type, np.uint64):
                return UInt64()
            elif is_np_dtype(pandas_type, np.uint32):
                return UInt32()
            elif is_np_dtype(pandas_type, np.uint16):
                return UInt16()
            elif is_np_dtype(pandas_type, np.uint8):
                return UInt8()
            raise TypeError
        if not workaround or pd.api.types.is_float_dtype(pandas_type):
            if not workaround or is_np_dtype(pandas_type, np.float64):
                return Float64()
            elif is_np_dtype(pandas_type, np.float32):
                return Float32()
            raise TypeError
        if pd.api.types.is_string_dtype(pandas_type):
            # We reserve the use of the object column for string.
            return String()
        if pd.api.types.is_bool_dtype(pandas_type):
            return Bool()
        if pd.api.types.is_datetime64_any_dtype(pandas_type):
            return Datetime()
        if pd.api.types.is_timedelta64_dtype(pandas_type):
            return Duration()
        # we don't know any decimal/time/null dtypes in pandas if column is not
        # arrow backed

        if pandas_type.name == "category":
            return Enum(*pandas_type.categories.to_list())

        raise TypeError

    @staticmethod
    def from_arrow(arrow_type) -> "Dtype":
        """Convert a PyArrow type to a Dtype."""
        import pyarrow as pa

        if pa.types.is_signed_integer(arrow_type):
            if pa.types.is_int64(arrow_type):
                return Int64()
            if pa.types.is_int32(arrow_type):
                return Int32()
            if pa.types.is_int16(arrow_type):
                return Int16()
            if pa.types.is_int8(arrow_type):
                return Int8()
            raise TypeError
        if pa.types.is_unsigned_integer(arrow_type):
            if pa.types.is_uint64(arrow_type):
                return UInt64()
            if pa.types.is_uint32(arrow_type):
                return UInt32()
            if pa.types.is_uint16(arrow_type):
                return UInt16()
            if pa.types.is_uint8(arrow_type):
                return UInt8()
            raise TypeError
        if pa.types.is_floating(arrow_type):
            if pa.types.is_float64(arrow_type):
                return Float64()
            if pa.types.is_float32(arrow_type):
                return Float32()
            if pa.types.is_float16(arrow_type):
                return Float32()
            raise TypeError
        if pa.types.is_decimal(arrow_type):
            # We don't recommend using Decimal in dataframes, but we support it.
            return Decimal(arrow_type.precision, arrow_type.scale)
        if pa.types.is_string(arrow_type):
            return String()
        if pa.types.is_boolean(arrow_type):
            return Bool()
        if pa.types.is_timestamp(arrow_type):
            return Datetime()
        if pa.types.is_date(arrow_type):
            return Date()
        if pa.types.is_time(arrow_type):
            return Time()
        if pa.types.is_duration(arrow_type):
            return Duration()
        if pa.types.is_null(arrow_type):
            return NullType()
        if pa.types.is_list(arrow_type):
            return List(Dtype.from_arrow(arrow_type.value_type))
        if pa.types.is_dictionary(arrow_type):
            raise RuntimeError(
                "Most likely this is an Enum type. But metadata about categories is "
                "only in the pyarrow field and not in the pyarrow dtype"
            )
        raise TypeError

    @staticmethod
    def from_polars(polars_type) -> "Dtype":
        """Convert a Polars type to a Dtype."""
        import polars as pl

        if isinstance(polars_type, pl.List):
            return List(Dtype.from_polars(polars_type.inner))
        if isinstance(polars_type, pl.Enum):
            return Enum(*polars_type.categories)
        if isinstance(polars_type, pl.Decimal):
            return Decimal(polars_type.precision, polars_type.scale)

        return {
            pl.Int64: Int64(),
            pl.Int32: Int32(),
            pl.Int16: Int16(),
            pl.Int8: Int8(),
            pl.UInt64: UInt64(),
            pl.UInt32: UInt32(),
            pl.UInt16: UInt16(),
            pl.UInt8: UInt8(),
            pl.Float64: Float64(),
            pl.Float32: Float32(),
            pl.Utf8: String(),
            pl.Boolean: Bool(),
            pl.Datetime: Datetime(),
            pl.Time: Time(),
            pl.Date: Date(),
            pl.Null: NullType(),
            pl.Duration: Duration(),
        }[polars_type.base_type()]

    def to_sql(self):
        """Convert this Dtype to a SQL type."""
        import sqlalchemy as sa

        return {
            Int(): sa.BigInteger(),  # we default to 64 bit
            Int8(): sa.SmallInteger(),
            Int16(): sa.SmallInteger(),
            Int32(): sa.Integer(),
            Int64(): sa.BigInteger(),
            UInt8(): sa.SmallInteger(),
            UInt16(): sa.Integer(),
            UInt32(): sa.BigInteger(),
            UInt64(): sa.BigInteger(),
            Float(): sa.Float(53),  # we default to 64 bit
            Float32(): sa.Float(24),
            Float64(): sa.Float(53),
            String(): sa.String(),
            Bool(): sa.Boolean(),
            Date(): sa.Date(),
            Time(): sa.Time(),
            Datetime(): sa.DateTime(),
            Duration(): sa.Interval(),
            NullType(): sa.types.NullType(),
        }[self]

    def to_pandas(self, backend: PandasBackend = PandasBackend.ARROW):
        """Convert this Dtype to a pandas type."""
        import pandas as pd

        if backend == PandasBackend.NUMPY:
            return self.to_pandas_nullable(backend)
        if backend == PandasBackend.ARROW:
            if isinstance(self, String) or isinstance(self, Enum):
                return pd.StringDtype(storage="pyarrow")
            return pd.ArrowDtype(self.to_arrow())

    def to_pandas_nullable(self, backend: PandasBackend = PandasBackend.ARROW):
        """Convert this Dtype to a pandas nullable type.

        Nullable can be either pandas extension types like StringDtype or ArrowDtype.

        Parameters
        ----------
        backend : PandasBackend, optional
            The pandas backend to use. Defaults to ``PandasBackend.ARROW``.
            If ``PandasBackend.NUMPY`` is selected, this method will attempt
            to return a NumPy-backed nullable pandas dtype. Note that
            Time, NullType, and List will raise a TypeError for the
            NUMPY backend as pandas doesn't have corresponding native
            nullable dtypes for these.
        """
        import pandas as pd

        if backend == PandasBackend.ARROW:
            return pd.ArrowDtype(self.to_arrow())

        # we don't want to produce object columns
        if isinstance(self, Time):
            raise TypeError("pandas doesn't have a native time dtype")
        if isinstance(self, NullType):
            raise TypeError("pandas doesn't have a native null dtype")
        if isinstance(self, List):
            raise TypeError("pandas doesn't have a native list dtype")

        if isinstance(self, Enum):
            return pd.CategoricalDtype(self.categories)
        if isinstance(self, String):
            return pd.StringDtype()  # max_length not needed for dataframes
        if isinstance(self, Decimal):
            # NumericDtype exists but is not used because fixpoint is more common
            # in SQL than in dataframes.
            return pd.Float64Dtype()
        if isinstance(self, List):
            # we don't want to produce object columns
            raise TypeError("pandas doesn't have a native list dtype")

        return {
            Int(): pd.Int64Dtype(),  # we default to 64 bit
            Int8(): pd.Int8Dtype(),
            Int16(): pd.Int16Dtype(),
            Int32(): pd.Int32Dtype(),
            Int64(): pd.Int64Dtype(),
            UInt8(): pd.UInt8Dtype(),
            UInt16(): pd.UInt16Dtype(),
            UInt32(): pd.UInt32Dtype(),
            UInt64(): pd.UInt64Dtype(),
            Float(): pd.Float64Dtype(),  # we default to 64 bit
            Float32(): pd.Float32Dtype(),
            Float64(): pd.Float64Dtype(),
            Bool(): pd.BooleanDtype(),
            Date(): "datetime64[s]",
            Datetime(): "datetime64[us]",
            Time(): "timedelta64[us]",
            Duration(): "timedelta64[us]",
        }[self]

    def to_arrow(self):
        """Convert this Dtype to a PyArrow type."""
        import pyarrow as pa

        return {
            Int(): pa.int64(),  # we default to 64 bit
            Int8(): pa.int8(),
            Int16(): pa.int16(),
            Int32(): pa.int32(),
            Int64(): pa.int64(),
            UInt8(): pa.uint8(),
            UInt16(): pa.uint16(),
            UInt32(): pa.uint32(),
            UInt64(): pa.uint64(),
            Float(): pa.float64(),  # we default to 64 bit
            Float32(): pa.float32(),
            Float64(): pa.float64(),
            String(): pa.string(),
            Bool(): pa.bool_(),
            Date(): pa.date32(),
            Time(): pa.time64("us"),
            Datetime(): pa.timestamp("us"),
            Duration(): pa.duration("us"),
            NullType(): pa.null(),
        }[self]

    def to_arrow_field(self, name: str, nullable: bool = True):
        """Convert this Dtype to a PyArrow Field."""
        import pyarrow as pa

        return pa.field(name, self.to_arrow(), nullable=nullable)

    def to_polars(self: "Dtype"):
        """Convert this Dtype to a Polars type."""
        import polars as pl

        return {
            Int(): pl.Int64,  # we default to 64 bit
            Int64(): pl.Int64,
            Int32(): pl.Int32,
            Int16(): pl.Int16,
            Int8(): pl.Int8,
            UInt64(): pl.UInt64,
            UInt32(): pl.UInt32,
            UInt16(): pl.UInt16,
            UInt8(): pl.UInt8,
            Float(): pl.Float64,  # we default to 64 bit
            Float64(): pl.Float64,
            Float32(): pl.Float32,
            Bool(): pl.Boolean,
            Datetime(): pl.Datetime("us"),
            Duration(): pl.Duration("us"),
            Time(): pl.Time,  # Polars uses nanoseconds since midnight
            Date(): pl.Date,
            NullType(): pl.Null,
        }[self]


class Float(Dtype):
    @classmethod
    def is_float(cls):
        return True


class Float64(Float): ...


class Float32(Float): ...


class Decimal(Float):
    def __init__(self, precision: int | None = None, scale: int | None = None):
        """
        Initialize a Decimal Dtype.

        Default is Decimal(31,10) which is the highest precision that works with DB2.
        If you like to save memory, Decimal(15,6) will get you quite far as well.

        :param precision: total number of digits in the number
            If not specified, it is assumed to be 31.
        :param scale: number of digits after the decimal point
            If not specified, it is assumed to be (precision//3+1).
        """
        self.precision = precision or 31
        self.scale = scale or (self.precision // 3 + 1)

    def __eq__(self, rhs):
        return isinstance(rhs, self.__class__) and self.precision == rhs.precision and self.scale == rhs.scale

    def __hash__(self):
        return hash((self.__class__.__name__, self.precision, self.scale))

    def __repr__(self):
        """Return a string representation of this dtype."""
        return f"{self.__class__.__name__}({self.precision}, {self.scale})"

    def to_sql(self):
        import sqlalchemy as sa

        return sa.Numeric(self.precision, self.scale)

    def to_polars(self):
        import polars as pl

        return pl.Decimal(self.precision, self.scale)

    def to_arrow(self):
        import pyarrow as pa

        if self.precision > 38:
            return pa.decimal256(self.precision, self.scale)
        elif self.precision > 18:
            return pa.decimal128(self.precision, self.scale)
        elif self.precision > 9:
            return pa.decimal64(self.precision, self.scale)
        else:
            return pa.decimal32(self.precision, self.scale)


class Int(Dtype):
    @classmethod
    def is_int(cls):
        return True


class Int64(Int): ...


class Int32(Int): ...


class Int16(Int): ...


class Int8(Int): ...


class UInt64(Int): ...


class UInt32(Int): ...


class UInt16(Int): ...


class UInt8(Int): ...


class String(Dtype):
    def __init__(self, max_length: int | None = None):
        """
        Initialize a String Dtype.

        :param max_length: maximum length of string
            This length will only be used for specifying fixed length strings in SQL.
            Thus, the meaning of characters vs. bytes is dependent on the SQL dialect.
        """
        self.max_length = max_length

    def __eq__(self, rhs):
        return isinstance(rhs, self.__class__) and self.max_length == rhs.max_length

    def __hash__(self):
        return hash((self.__class__.__name__, self.max_length))

    def __repr__(self):
        """Return a string representation of this dtype."""
        return f"{self.__class__.__name__}({self.max_length})"

    def to_sql(self):
        """Convert this Dtype to a SQL type."""
        import sqlalchemy as sa

        return sa.String(length=self.max_length)

    def to_polars(self):
        import polars as pl

        return pl.Utf8

    def to_arrow(self):
        import pyarrow as pa

        return pa.string()


class Bool(Dtype): ...


class Datetime(Dtype): ...


class Date(Dtype): ...


class Time(Dtype): ...


class Duration(Dtype): ...


class NullType(Dtype): ...


class List(Dtype):
    def __init__(self, inner: "Dtype"):
        self.inner = inner

    def __eq__(self, rhs):
        return isinstance(rhs, List) and self.inner == rhs.inner

    def __hash__(self):
        return hash((self.__class__.__name__, hash(self.inner)))

    def __repr__(self):
        """Return a string representation of this dtype."""
        return f"{self.__class__.__name__}[{self.inner}]"

    def to_sql(self):
        import sqlalchemy as sa

        return sa.ARRAY(self.inner.to_sql())

    def to_polars(self):
        import polars as pl

        return pl.List(self.inner.to_polars())

    def to_arrow(self):
        import pyarrow as pa

        return pa.list_(self.inner.to_arrow())


class Enum(String):
    def __init__(self, *categories: str):
        if not all(isinstance(c, str) for c in categories):
            raise TypeError("arguments for `Enum` must have type `str`")
        self.categories = list(categories)
        self.max_length = max([len(c) for c in categories]) if len(categories) > 0 else None

    def __eq__(self, rhs):
        return isinstance(rhs, Enum) and self.categories == rhs.categories

    def __hash__(self):
        return hash((self.__class__.__name__, tuple(self.categories)))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}[{', '.join(repr(c) for c in self.categories)}]"

    def to_polars(self):
        import polars as pl

        return pl.Enum(self.categories)

    def to_arrow(self):
        import pyarrow as pa

        # enum categories can only be maintained in pyarrow field (see to_arrow_field)
        return pa.string()

    def to_arrow_field(self, name: str, nullable: bool = True):
        """Convert this Dtype to a PyArrow Field."""
        import pyarrow as pa

        # try to mimic what polars does
        return pa.field(
            name,
            pa.dictionary(pa.uint32(), pa.large_string()),
            nullable=nullable,
            metadata={
                # the key might change with polars versions
                "_PL_ENUM_VALUES2": "".join([f"{len(c)};{c}" for c in self.categories])
            },
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import dis
import inspect
from enum import Enum
from typing import Any


class Operation(Enum):
    OBJECT = 0
    GETATTR = 10
    SETATTR = 11
    DELATTR = 12
    GETITEM = 20
    SETITEM = 21
    DELITEM = 22
    CALL = 50
    GET = 60
    BOOL = 70

    def __repr__(self):
        return self.name


class ComputationTracer:
    proxy_type: type["ComputationTracerProxy"]

    def __init__(self):
        self.trace = []
        self.patcher = MonkeyPatcher()
        self.did_exit = False
        self.proxy_type = ComputationTracerProxy

    def create_proxy(self, identifier=None):
        return self._get_proxy((Operation.OBJECT, identifier))

    def __enter__(self):
        self._monkey_patch()
        # clear trace already filled during patching (modules may issue calls during
        # initialization)
        self.trace = []
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.patcher.undo()
        self.did_exit = True

    def _get_proxy(self, computation: tuple):
        idx = len(self.trace)
        self._add_computation(computation)
        return self.proxy_type(self, idx)

    def _add_computation(self, computation: tuple):
        if not self.did_exit:
            from pydiverse.common.util.deep_map import deep_map

            computation = deep_map(computation, self._computation_mapper)
            self.trace.append(computation)
        else:
            raise RuntimeError("Can't modify ComputationTrace after exiting the context.")

    @staticmethod
    def _computation_mapper(x):
        if isinstance(x, ComputationTracerProxy):
            return ComputationTraceRef(x)

        if inspect.isfunction(x):
            bytecode = dis.Bytecode(x)
            return (
                Operation.OBJECT,
                "BYTECODE",
                tuple((instr.opcode, instr.argval) for instr in bytecode),
            )

        return x

    def _monkey_patch(self): ...

    def trace_hash(self) -> str:
        try:
            from dask.base import tokenize
        except ModuleNotFoundError:
            raise ModuleNotFoundError("module dask is required to use computation_tracing.") from None

        return tokenize(self.trace)


class ComputationTracerProxy:
    def __init__(self, tracer: ComputationTracer, identifier: int | str):
        object.__setattr__(self, "_computation_tracer_", tracer)
        object.__setattr__(self, "_computation_tracer_id_", identifier)

    def __getattribute__(self, item):
        if item in ("__class__", "__module__"):
            return object.__getattribute__(self, item)

        tracer = _get_tracer(self)
        return tracer._get_proxy((Operation.GETATTR, self, item))

    def __setattr__(self, key, value):
        tracer = _get_tracer(self)
        tracer._add_computation((Operation.SETATTR, self, key, value))

    def __delattr__(self, key):
        tracer = _get_tracer(self)
        tracer._add_computation((Operation.DELATTR, self, key))

    def __getitem__(self, item):
        tracer = _get_tracer(self)
        return tracer._get_proxy((Operation.GETITEM, self, item))

    def __setitem__(self, key, value):
        tracer = _get_tracer(self)
        tracer._add_computation((Operation.SETITEM, self, key, value))

    def __delitem__(self, key):
        tracer = _get_tracer(self)
        tracer._add_computation((Operation.DELITEM, self, key))

    def __call__(self, *args, **kwargs):
        tracer = _get_tracer(self)
        return tracer._get_proxy((Operation.CALL, self, args, kwargs))

    def __get__(self, instance, owner):
        tracer = _get_tracer(self)
        return tracer._get_proxy((Operation.GET, instance, self))

    def __bool__(self):
        tracer = _get_tracer(self)
        return tracer._get_proxy((Operation.BOOL, self))

    def __iter__(self):
        raise RuntimeError("__iter__ is not supported by ComputationTracerProxy")

    def __contains__(self, item):
        raise RuntimeError("__contains__ is not supported by ComputationTracerProxy")

    def __len__(self):
        raise RuntimeError("__len__ is not supported by ComputationTracerProxy")


def _get_tracer(proxy: ComputationTracerProxy) -> ComputationTracer:
    return object.__getattribute__(proxy, "_computation_tracer_")


__supported_dunder = {
    "__add__",
    "__radd__",
    "__sub__",
    "__rsub__",
    "__mul__",
    "__rmul__",
    "__truediv__",
    "__rtruediv__",
    "__floordiv__",
    "__rfloordiv__",
    "__pow__",
    "__rpow__",
    "__mod__",
    "__rmod__",
    "__round__",
    "__pos__",
    "__neg__",
    "__abs__",
    "__and__",
    "__rand__",
    "__or__",
    "__ror__",
    "__xor__",
    "__rxor__",
    "__invert__",
    "__lt__",
    "__le__",
    "__eq__",
    "__ne__",
    "__gt__",
    "__ge__",
    "__copy__",
    "__deepcopy__",
}


def __create_dunder(name):
    def dunder(self, *args):
        return getattr(self, name)(self, *args)

    return dunder


for dunder_ in __supported_dunder:
    setattr(ComputationTracerProxy, dunder_, __create_dunder(dunder_))


class ComputationTraceRef:
    __slots__ = ("id",)

    def __init__(self, proxy: ComputationTracerProxy):
        self.id = object.__getattribute__(proxy, "_computation_tracer_id_")

    def __str__(self):
        return f"ComputationTraceRef<{self.id}>"

    def __repr__(self):
        return f"ComputationTraceRef<{self.id}>"

    def __dask_tokenize__(self):
        return "ComputationTraceRef", self.id


class MonkeyPatcher:
    """Monkey Patching class inspired by pytest's MonkeyPatch class"""

    def __init__(self):
        self._setattr: list[tuple[object, str, Any]] = []

    def patch_attr(self, obj: object, name: str, value: Any):
        old_value = getattr(obj, name)

        # avoid class descriptors like staticmethod / classmethod
        if inspect.isclass(obj):
            old_value = obj.__dict__[name]

        setattr(obj, name, value)
        self._setattr.append((obj, name, old_value))

    def undo(self):
        for obj, name, value in reversed(self._setattr):
            setattr(obj, name, value)
        self._setattr.clear()


def fully_qualified_name(obj):
    if type(obj).__name__ == "builtin_function_or_method":
        if obj.__module__ is not None:
            module = obj.__module__
        else:
            if inspect.isclass(obj.__self__):
                module = obj.__self__.__module__
            else:
                module = obj.__self__.__class__.__module__
        return f"{module}.{obj.__qualname__}"

    if type(obj).__name__ == "function":
        if hasattr(obj, "__wrapped__"):
            qualname = obj.__wrapped__.__qualname__
        else:
            qualname = obj.__qualname__
        return f"{obj.__module__}.{qualname}"

    if type(obj).__name__ in (
        "member_descriptor",
        "method_descriptor",
        "wrapper_descriptor",
    ):
        return f"{obj.__objclass__.__module__}.{obj.__qualname__}"

    if type(obj).__name__ == "method":
        if inspect.isclass(obj.__self__):
            cls = obj.__self__.__qualname__
        else:
            cls = obj.__self__.__class__.__qualname__
        return f"{obj.__self__.__module__}.{cls}.{obj.__name__}"

    if type(obj).__name__ == "method-wrapper":
        return f"{fully_qualified_name(obj.__self__)}.{obj.__name__}"

    if type(obj).__name__ == "module":
        return obj.__name__

    if type(obj).__name__ == "property":
        return f"{obj.fget.__module__}.{obj.fget.__qualname__}"

    if inspect.isclass(obj):
        return f"{obj.__module__}.{obj.__qualname__}"

    return f"{obj.__class__.__module__}.{obj.__class__.__qualname__}"


def patch(tracer: ComputationTracer, target: object, name: str):
    if isinstance(target, ComputationTracerProxy):
        return

    val = getattr(target, name)

    if isinstance(val, type):
        return patch_type(tracer, target, name)
    if inspect.isfunction(val):
        return patch_function(tracer, target, name)
    if isinstance(val, object):
        return patch_value(tracer, target, name)

    raise RuntimeError


def patch_type(tracer: ComputationTracer, target: object, name: str):
    val = getattr(target, name)
    full_name = fully_qualified_name(val)

    dunder_to_patch = (
        "__getattr__",
        "__setattr__",
        "__delattr__",
        "__getitem__",
        "__setitem__",
        "__delitem__",
        "__call__",
        "__repr__",
        "__str__",
    )

    for key, _value in object.__getattribute__(val, "__dict__").items():
        if key.startswith("__"):
            if key in dunder_to_patch:
                try:
                    tracer.patcher.patch_attr(val, key, tracer.proxy_type(tracer, full_name + "." + key))
                except AttributeError:
                    pass
            continue
        else:
            tracer.patcher.patch_attr(val, key, tracer.proxy_type(tracer, full_name + "." + key))

    tracer.patcher.patch_attr(target, name, tracer.proxy_type(tracer, full_name))


def patch_function(tracer: ComputationTracer, target: object, name: str):
    val = getattr(target, name)
    full_name = fully_qualified_name(val)
    tracer.patcher.patch_attr(target, name, tracer.proxy_type(tracer, full_name))


def patch_value(tracer: ComputationTracer, target: object, name: str):
    full_name = fully_qualified_name(target) + "." + name
    tracer.patcher.patch_attr(target, name, tracer.proxy_type(tracer, full_name))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import base64
import hashlib
import types
import warnings

try:
    import polars as pl
except ImportError:
    pl = types.ModuleType("pl")
    pl.DataFrame = None

try:
    import pyarrow as pa
    import pyarrow.ipc as ipc
    import pyarrow.types as pat
except ImportError:
    pa = types.ModuleType("pa")
    pa.DataType = None
    pa.Table = None
    pa.Schema = None
    pat = types.ModuleType("pat")
    ipc = types.ModuleType("ipc")


def stable_hash(*args: str) -> str:
    """Compute a hash over a set of strings

    :param args: Some strings from which to compute the cache key
    :return: A sha256 base32 digest, trimmed to 20 char length
    """

    combined_hash = hashlib.sha256(b"PYDIVERSE")
    for arg in args:
        arg_bytes = str(arg).encode("utf8")
        arg_bytes_len = len(arg_bytes).to_bytes(length=8, byteorder="big")

        combined_hash.update(arg_bytes_len)
        combined_hash.update(arg_bytes)

    # Only take first 20 characters of base32 digest (100 bits). This
    # provides 50 bits of collision resistance, which is more than enough.
    # To illustrate: If you were to generate 1k hashes per second,
    # you still would have to wait over 800k years until you encounter
    # a collision.

    # NOTE: Can't use base64 because it contains lower and upper case
    #       letters; identifiers in pipedag are all lowercase
    hash_digest = combined_hash.digest()
    hash_str = base64.b32encode(hash_digest).decode("ascii").lower()
    return hash_str[:20]


def _hash_polars_dataframe(df: pl.DataFrame, use_init_repr=False) -> str:
    """
    Compute a stable hash for a polars DataFrame. This function is meant as a backup for
    the pyarrow hashing function in case it breaks.

    The function may fail for some known causes:
    * In old polars versions, hash_rows() does not work if there are list columns
      https://github.com/pola-rs/polars/issues/24121
    * Hashing categorical and enum columns may lead to unstable results depending on the internal dictionary encoding.
    * E.g.
    ```python
    import polars as pl
    df_c = pl.DataFrame(dict(x=["c", "b", "c"]), schema=dict(x=pl.Enum(["b", "c"])))
    _hash_polars_dataframe(df_c)
    ```
    and
    ```python
    import polars as pl
    df_a = pl.DataFrame(dict(x=["apple", "banana", "apple"]), schema=dict(x=pl.Enum(["apple", "banana"])))
    df_c = pl.DataFrame(dict(x=["c", "b", "c"]), schema=dict(x=pl.Enum(["b", "c"])))
    _hash_polars_dataframe(df_c)
    ```
    will give different results because the internal dictionary encoding is different. Same applies
    to the Enum type replaced by Categorical.
    See also https://github.com/pola-rs/polars/issues/14829
    """
    if not use_init_repr:
        try:
            schema_hash = stable_hash(repr(df.schema))
            if df.is_empty():
                content_hash = "empty"
            else:
                content_hash = str(
                    df.hash_rows()  # We get a Series of hashes, one for each row
                    # Since polars only hashes rows, we need to implode the Series into
                    # a single row to get a single hash
                    .implode()
                    .hash()
                    .item()
                )
            return "0" + stable_hash(schema_hash, content_hash)
        except Exception:
            warnings.warn(
                "Failed to compute hash for polars DataFrame in fast way. Falling back to to_init_repr() method.",
                stacklevel=1,
            )

    # fallback to to_init_repr string representation
    return "1" + stable_hash(df.to_init_repr(len(df)))


# ---------- type helpers ----------


def _without_dict_type(t: pa.DataType) -> pa.DataType:
    """Return a type like t but with all Dictionary types replaced by their value types (recursively)."""
    if pat.is_dictionary(t):
        return t.value_type
    if pat.is_struct(t):
        new_fields = [pa.field(f.name, _without_dict_type(f.type), nullable=f.nullable) for f in t]
        return pa.struct(new_fields)
    if pat.is_list(t):
        return pa.list_(_without_dict_type(t.value_type))
    if pat.is_large_list(t):
        return pa.large_list(_without_dict_type(t.value_type))
    if pat.is_fixed_size_list(t):
        return pa.list_(_without_dict_type(t.value_type), t.list_size)
    if pat.is_map(t):
        return pa.map_(
            _without_dict_type(t.key_type),
            _without_dict_type(t.item_type),
            keys_sorted=t.keys_sorted,
        )
    # (We leave unions and other rare complex types as-is; they typically don't carry dictionaries)
    return t


def _strip_field_metadata(schema: pa.Schema) -> pa.Schema:
    """Drop field-level metadata to avoid non-semantic differences."""
    fields = [pa.field(f.name, f.type, nullable=f.nullable) for f in schema]
    return pa.schema(fields, metadata=None)


def _dictionary_free_schema(schema: pa.Schema) -> pa.Schema:
    """Make a schema with identical field names/nullable flags, but no dictionary types (recursively)."""
    fields = [pa.field(f.name, _without_dict_type(f.type), nullable=f.nullable) for f in schema]
    return pa.schema(fields)  # no schema metadata


def _as_arrow_table(obj) -> pa.Table:
    """Accept pyarrow Table, Polars DF, pandas DF and return a pa.Table."""
    if isinstance(obj, pa.Table):
        return obj

    # Polars DataFrame
    if pl is not None and isinstance(obj, pl.DataFrame):
        return obj.to_arrow()

    # Pandas DataFrame
    import pandas as pd

    if isinstance(obj, pd.DataFrame):
        return pa.Table.from_pandas(obj, preserve_index=True)

    raise TypeError(f"Unsupported input type: {type(obj)!r}")


def _hash_schema(obj) -> str:
    """Compute a stable hash for the schema of a pyarrow Table, Polars DataFrame, or pandas DataFrame."""
    if isinstance(obj, pa.Table):
        return stable_hash(repr(obj.schema))
    if pl is not None and isinstance(obj, pl.DataFrame):
        return stable_hash(repr(obj.schema))
    import pandas as pd

    if isinstance(obj, pd.DataFrame):
        return stable_dataframe_hash(pa.Table.from_pandas(pd.DataFrame(obj.dtypes.astype("string[pyarrow]"))))
    raise TypeError(f"Unsupported input type: {type(obj)!r}")


def stable_dataframe_hash(
    df,
    *,
    strip_schema_metadata: bool = True,
    drop_field_metadata: bool = True,
) -> str:
    """
    Compute a stable, cross-library SHA-256 of a full table.

    Accepts: Polars DataFrame, PyArrow Table, pandas DataFrame.

    What makes it stable:
      • combine_chunks()
      • recursively decode dictionary/categorical encodings via cast to a dict-free schema
      • strip schema + field metadata
      • serialize as a single-batch Arrow IPC stream with pinned options and hash the bytes
    """
    # 1) Convert to Arrow and canonicalize chunking
    t = _as_arrow_table(df).combine_chunks()

    # 2) Drop non-semantic metadata (schema + field-level)
    if strip_schema_metadata:
        t = t.replace_schema_metadata(None)
    if drop_field_metadata:
        t = pa.Table.from_arrays(list(t.columns), schema=_strip_field_metadata(t.schema))

    # 3) Hash schema separately, since we replace dictionaries in the next step
    # If we don't have a pyarrow Table, we use the original df to compute the schema hash
    # because some dtype information may be lost in the conversion to pyarrow.
    # For example, after converting from pandas to pyarrow, we no longer know if a column
    # was arrow or numpy backed.
    schema_hash = _hash_schema(df) if not isinstance(df, pa.Table) else _hash_schema(t)

    # 4) Decode all dictionary encodings (including nested) by casting to a dict-free schema
    dict_free = _dictionary_free_schema(t.schema)
    if dict_free != t.schema:
        t = t.cast(dict_free)

    # 5) Deterministic IPC serialization (single batch, no compression, explicit metadata version)
    sink = pa.BufferOutputStream()
    write_opts = ipc.IpcWriteOptions(metadata_version=ipc.MetadataVersion.V5, compression=None)
    with ipc.RecordBatchStreamWriter(sink, t.schema, options=write_opts) as writer:
        writer.write_table(t, max_chunksize=t.num_rows)

    # 6) Hash subtleties of pandas DataFrame object type columns that are lost when converting
    # to pyarrow, e.g. columns mixing dates and datetimes, by hashing their CSV representation.
    # See https://github.com/apache/arrow/issues/41896 for an example of such a conversion issue.
    if not isinstance(df, (pa.Table, pl.DataFrame)):
        import pandas as pd

        assert isinstance(df, pd.DataFrame), f"Dataframe of type {type(df)} not supported"

        df_object = df.select_dtypes(include=["object"])
        csv = df_object.to_csv()
    else:
        csv = ""

    return stable_hash(schema_hash, hashlib.sha256(sink.getvalue().to_pybytes()).hexdigest(), csv)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

"""Generic deep map or mutation operations.

Heavily inspired by the builtin copy module of python:
https://github.com/python/cpython/blob/main/Lib/copy.py
"""

from collections.abc import Callable

from .computation_tracing import fully_qualified_name
from .import_ import load_object

_nil = []


def deep_map(x, fn: Callable, memo=None):
    if memo is None:
        memo = {}

    d = id(x)
    y = memo.get(d, _nil)
    if y is not _nil:
        return y

    cls = type(x)

    if cls == list:  # noqa: E721
        y = _deep_map_list(x, fn, memo)
    elif cls == tuple:  # noqa: E721
        y = _deep_map_tuple(x, fn, memo)
    elif cls == dict:  # noqa: E721
        y = _deep_map_dict(x, fn, memo)
    elif hasattr(cls, "__dataclass_fields__"):
        # reconstruct data classes
        y = load_object(
            {
                "class": fully_qualified_name(cls),
                "args": _deep_map_dict(x.__dict__, fn, memo),
            }
        )
    else:
        y = x

    y = fn(y)

    # If is its own copy, don't memoize.
    if y is not x:
        memo[d] = y
        _keep_alive(x, memo)  # Make sure x lives at least as long as d

    return y


def _deep_map_list(x, fn, memo):
    y = []
    append = y.append
    for a in x:
        append(deep_map(a, fn, memo))
    return fn(y)


def _deep_map_tuple(x, fn, memo):
    y = [deep_map(a, fn, memo) for a in x]
    # We're not going to put the tuple in the memo, but it's still important we
    # check for it, in case the tuple contains recursive mutable structures.
    try:
        return memo[id(x)]
    except KeyError:
        pass
    for k, j in zip(x, y, strict=False):
        if k is not j:
            y = tuple(y)
            break
    else:
        y = x
    return fn(y)


def _deep_map_dict(x, fn, memo):
    y = {}
    memo[id(x)] = y
    for key, value in x.items():
        y[deep_map(key, fn, memo)] = deep_map(value, fn, memo)
    return fn(y)


def _keep_alive(x, memo):
    """Keeps a reference to the object x in the memo.
    Because we remember objects by their id, we have
    to assure that possibly temporary objects are kept
    alive by referencing them.
    We store a reference at the id of the memo, which should
    normally not be used unless someone tries to deepcopy
    the memo itself...
    """
    try:
        memo[id(memo)].append(x)
    except KeyError:
        # aha, this is the first one :-)
        memo[id(memo)] = [x]



# Copyright (c) QuantCo and pydiverse contributors 2025-2026
# SPDX-License-Identifier: BSD-3-Clause
import logging
import sys
import textwrap
import types
from collections.abc import Generator
from contextlib import contextmanager
from io import StringIO

try:
    import structlog
    from structlog import configure, get_config
    from structlog.testing import LogCapture
    from structlog.typing import EventDict, WrappedLogger

    structlog_installed = True
except ImportError:
    structlog = types.ModuleType("structlog")

    class dev:
        class ConsoleRenderer:
            pass

        plain_traceback = None

    structlog.dev = dev
    structlog_installed = False
    configure, get_config = None, None
    LogCapture = None
    EventDict, WrappedLogger = None, None


class PydiverseConsoleRenderer(structlog.dev.ConsoleRenderer):
    """
    Custom subclass of the structlog ConsoleRenderer that allows rendering
    specific values in the event dict on separate lines.
    """

    def __init__(self, *args, **kwargs):
        self._render_keys = kwargs.pop("render_keys", [])
        super().__init__(*args, **kwargs)

    def __call__(self, logger: WrappedLogger, name: str, event_dict: EventDict):
        render_objects = {}
        for key in self._render_keys:
            obj = event_dict.pop(key, None)
            if obj is not None:
                render_objects[key] = obj

        result = super().__call__(logger, name, event_dict)
        sio = StringIO()
        sio.write(result)

        for key, obj in render_objects.items():
            string_rep = str(obj)
            sio.write(
                "\n"
                + "    ["
                + self._styles.kv_key
                + key
                + self._styles.reset
                + "]"
                + "\n"
                + textwrap.indent(string_rep, prefix="    " + self._styles.kv_value)
                + self._styles.reset
            )

        return sio.getvalue()


def setup_logging(
    log_level=logging.INFO,
    log_stream=None,
    timestamp_format="%Y-%m-%d %H:%M:%S.%f",
    exception_formatter=structlog.dev.plain_traceback,
):
    """Configures structlog and logging with sane defaults."""

    if log_stream is None:
        log_stream = sys.stderr

    # Stdlib: decide where logs go here (handlers), not in structlog’s LoggerFactory
    logging.basicConfig(
        level=log_level,
        handlers=[logging.StreamHandler(log_stream)],
        format="%(message)s",
    )

    if structlog_installed:
        # --- Final renderer used for BOTH stdlib and structlog events ---
        renderer = PydiverseConsoleRenderer(
            render_keys=["query", "table_obj", "task", "table", "detail"], exception_formatter=exception_formatter
        )

        # --- ProcessorFormatter wires stdlib logging -> structlog processors ---
        formatter = structlog.stdlib.ProcessorFormatter(
            # This is the final step for BOTH paths (stdlib & structlog):
            processor=renderer,
            # Processors applied to *foreign* (stdlib logging) records before 'processor'
            foreign_pre_chain=[
                structlog.contextvars.merge_contextvars,
                structlog.processors.StackInfoRenderer(),
                structlog.dev.set_exc_info,
                structlog.stdlib.add_log_level,
                # structlog.stdlib.filter_by_level,  # for logging, it won't have access to the actual logger
                structlog.processors.TimeStamper(timestamp_format),
            ],
        )

        # --- Stdlib handler (pytest caplog will capture this) ---
        handler = logging.StreamHandler(log_stream)
        handler.setFormatter(formatter)

        # It’s often safer to set handlers explicitly than to rely on basicConfig()’s global state
        root = logging.getLogger()
        root.handlers[:] = [handler]
        root.setLevel(log_level)

        # --- structlog side: end with wrap_for_formatter so ProcessorFormatter runs the renderer ---
        structlog.configure(
            processors=[
                structlog.contextvars.merge_contextvars,
                structlog.processors.StackInfoRenderer(),
                structlog.dev.set_exc_info,
                structlog.stdlib.add_log_level,
                structlog.stdlib.filter_by_level,  # consult stdlib level
                structlog.processors.TimeStamper(timestamp_format),
                # Hand off to logging's ProcessorFormatter (which will call `renderer`)
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            ],
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )


@contextmanager
def log_level(level: int, logger_name: str | None = None):
    lg = logging.getLogger(logger_name)  # None => root
    old = lg.level
    try:
        lg.setLevel(level)
        yield
    finally:
        lg.setLevel(old)


@contextmanager
def capture_logs() -> Generator[list[EventDict], None, None]:
    """
    Context manager that appends all logging statements to its yielded list
    while it is active. Disables all configured processors for the duration
    of the context manager.

    Attention: this is **not** thread-safe!

    <This is derived from structlog.testing.capture_logs which is MIT licensed>
    """
    assert structlog_installed, "please install structlog to capture logs with this function"

    cap = LogCapture()
    # Modify `_Configuration.default_processors` set via `configure` but always
    # keep the list instance intact to not break references held by bound
    # loggers.
    processors = get_config()["processors"]
    old_processors = processors.copy()
    try:
        # clear processors list and use LogCapture for testing
        processors.clear()
        processors.append(structlog.stdlib.add_log_level)
        processors.append(structlog.stdlib.filter_by_level)
        processors.append(cap)
        configure(processors=processors)
        yield cap.entries
    finally:
        # remove LogCapture and restore original processors
        processors.clear()
        processors.extend(old_processors)
        configure(processors=processors)
